%nolist
include list_def.inc
include psp_bda.inc
%list

debug = false

version_nr	equ	'1.5'

;-----------------------------------------------------------------------------
;
; Clock driver that make's DOS always read the AT Real Time Clock for date and
; time functions. It functions with a precision of 1/100 of a second. It uses
; the Programmable Interrupt Timer for that.
;
; Millennium proof.
;
; The improved precision (the normal DOS clock driver works with a precision
; of 1/18 of a second) can be handy for measuring time differences, and thus
; for testing the speed of some programs.
;
; Version history
;
; pre 1.0	Been using it with no problems for a few months.
;
; 1.0  25-08-98	Added more documentation, a help function and some exta tests
;		for a functioning RTC. No alterations in the resident part of
;		the device driver.
;
; 1.1  28-08-98	Added a patch for a BIOS with a millennium bug. And a test for
;		a PS/2, used to get the century register number of the RTC
;		right.
;
; 1.2  30-09-98 Removed a bug in the adjusting down code.
;		And now at startup we'll only adjust up. We're sort of just
;		waiting till the seconds change, than we set our count to 0 by
;		adjusting up.
;		A few very minor alterations in the resident code.
;
; 1.3  20-10-98 Added the possibility of installing the driver from the prompt.
;		It is done by an exe-part that is glued on at the end of the
;		file. This exe-part mimics the DOS install procedure, leaving
;		the installing of the interrupt vectors to this part of the
;		program. The linking in the device driver chain afterwards is
;		handled again by the exe-part.
;		Now if the PIT has to be reread because of a change of status,
;		the time is reread too, hopefully avoiding unwanted
;		interference.
;		Now the seconds are read directly from the RTC.
;		Added more debug code. Precision still isn't as good as I
;		expect. Irregularities still give that adjustments are
;		frequently made.
;		Added possibility to uninstall (also makes debugging easier).
;
;		Although it can be loaded as a 'TSR', in reality it isn't.
;		It's still a device driver. That means you can't uninstall it
;		using the MARK and RELEASE programs of TurboPower Software's
;		TSR Utilities, or programs that are alike.
;
; 1.4  06-11-98	Added default check and fix for 19-20 century roll-over at
;		startup. Simplified the fix of the millenium bug. Normally the
;		bug in the BIOS consists of the RTC failing to do the century
;		roll-over. We make sure reading gives the correct century.
;		Removed the fix millenium bug option. It's no longer needed.
;		The simplified version is always installed and probably does
;		the job  well enough
;
; 1.5  15-04-99 In Windows acuracy is not so good (In Windows 3.1 enhanced
;		mode the status of the PIT wasn't put through OK, and the
;		timer tick is put through too late), so we in Windows only use
;		a normal timer counter. Using the windows Virtual Timer Device
;		we could get a high accuracy again. But I don't see the point
;		of doing that, because Windows Time Slicing makes MS-DOS time
;		differences variable.
;
;-----------------------------------------------------------------------------

locals

instrc	equ	2			;offsets in device request header
status	equ	3
end_adr	equ	14
bfr_adr	equ	14
number	equ	18
cmdline	equ	18

.286C					;it's for an AT

model		tiny

codeseg
	assume	cs:@code, ds:nothing, es:nothing, ss:nothing
	org	0

next_d	dd	-1			;device header
	dw	8008h			;clock device
	dw	strat
	dw	intr
	db	'CLOCK$  '		;clock device

mon_tbl	db	31			;months table
febr	db	28
	db	31,30,31,30,31,31,30,31,30,31

strat	proc	far			;strategy routine
	mov	[db_ptr_o], bx
	mov	[db_ptr_s], es
	ret
strat	endp

intr	proc	far			;interrupt routine
pusha_instr equ	this byte		;wil become pusha
push_ds_pushf equ this word+1		;	    push  ds
	jmp	init			;	    pushf
	cld				; after installation
	push	'NN'
db_ptr_s equ this word-2		;place for segment of device request
	pop	ds			; header
	mov	si, 'NN'
db_ptr_o equ this word-2		;place for offset of device request
	lodsw				; header
	lodsb
	cmp	al, 8			;write command?
	je	write
	cmp	al, 4			;read command?
	mov	ax, 8103h		;presume unknown command
	jne	intr_end

;-----------------------------------------------------------------------------
;
; Request to read time
;
; Length of the structure used to read and write is 6 bytes
;
; 0 word days since 1-1-1980		;theoretically OK upto the year 2159
; 2 byte minutes
; 3 byte hours
; 4 byte 1/100's of seconds
; 5 byte seconds
;
;-----------------------------------------------------------------------------

read	proc				;reading time routine
	push	es
	les	di, [si+bfr_adr-3]	;point to destination

	mov	ah, 4			;read date
	int	1Ah
	xchg	ax, cx
	call	bcd_bin			;convert to binary
	sub	ah, 19			;get years since 1900
	aad	100
	sub	al, 80			;get years since 1980
	xchg	ax, cx
	xchg	ax, dx
	call	bcd_bin			;convert days and months to binary
	mov	dx, ax

	cbw				;start count with days
	call	febr_adj		;are we in a leap year?

	jmp	@@years_entry
@@years:test	cl, 3
	jne	@@no_lpy
	stc
@@no_lpy:
	adc	ax, 365
@@years_entry:
	dec	cx			;count days in the past years
	jns	@@years

@@months:				;count days in the past months
	mov	bx, offset mon_tbl
	jmp	@@months_entry
@@months_loop:
	add	al, cs:[bx]
	adc	ah, 0
	inc	bx
@@months_entry:
	dec	dh
	jnz	@@months_loop

@@ready:dec	ax			;1 too many
	stosw				;store it
	mov	ah, 2			;get time
	int	1Ah
	xchg	ax, cx			;minute & hour
	call	bcd_bin			;convert to binary
	stosw				;store hours and minutes
	call	time_read_and_sync	;read secs and synchronise
	stosw				;store seconds and 1/100's
	pop	es
intr_accomplished:
	mov	byte ptr [si+number-3], 6 ;set nr of bytes to 6
	mov	ax, 0100h		;signal ready
intr_end:
	mov	[si], ax
	popf
	pop	ds
	popa
	retf
read	endp
intr	endp

;-----------------------------------------------------------------------------
;
; Request to write time
;
;-----------------------------------------------------------------------------

write	proc	near			;writing time routine
	push	ds si
	lds	si, [si+bfr_adr-3]	;point to information
	mov	ax, [si]		;get days

	mov	cl, 79			;start with 1980
@@years:inc	cx			;count years past
	mov	bx, 365
	test	cl, 3
	jne	@@no_lpy
	inc	bx
@@no_lpy:
	sub	ax, bx
	jnb	@@years
	add	ax, bx			;1 year too many

@@months:
	call	febr_adj
	xor	bx, bx			;assume january
	cwd				;dh:= 0
@@months_loop:				;count months past
	inc	bx
	mov	dl, cs:[mon_tbl-next_d-1+bx]
	sub	ax, dx
	jnc	@@months_loop
	adc	ax, dx			;1 month too many, 1 day extra

;	mov	ah, bl			;day already in al
;	call	bin_bcd
	call	bin_bcd_direct		;convert to BCD
	xchg	dx, ax
	xchg	ax, cx
	aad	100			;get century right
	add	ah, 19
	call	bin_bcd			;convert to BCD
	xchg	ax, cx
	mov	ah, 5
	int	1Ah			;set date through BIOS

	lodsw				;get days again
	inc	ax			;we set the day of the week too
	inc	ax
	xor	dx, dx
	mov	cx, 7
	div	cx
	inc	dx
	mov	al, 6
	mov	ah, dl
	out	70h, ax			;set day of week directly in the RTC

	lodsw				;minute & hour
	call	bin_bcd			;convert to BCD
	xchg	cx, ax
	lodsw				;seconds, we discard the 1/100's
	xor	al, al			;no daylight savings time
	call	bin_bcd			;convert to BCD
	xchg	dx, ax
	mov	ah, 3
	int	1Ah			;set time through BIOS

	pop	si ds
	jmp	intr_accomplished	;we're ready
write	endp

febr_adj proc	near			;adjusts length of februari for leap
	mov	bl, 28			; years, year should be in cl
	test	cl, 3
	jne	@@no_lpm
	inc	bx
@@no_lpm:
	mov	[febr], bl
	ret
febr_adj endp

bin_bcd	proc	near			;converts ax from binary to packed bcd
	mov	bl, ah			; numbers, uses bx
bin_bcd_direct:
	aam				;convert
	aad	16			;put numbers together in al
	xchg	bx, ax
	aam
	aad	16
	jmp	bcd_proc_ret
bin_bcd	endp

bcd_bin	proc	near			;converts ax from packed bcd numbers
	mov	bl, ah			; to binary, uses bl
	aam	16			;separate numbers
	aad				;convert
	xchg	bx, ax
bcd_bin_half:
	aam	16
	aad
bcd_proc_ret:
	mov	bh, al
	xchg	ax, bx
	ret
bcd_bin	endp

;-----------------------------------------------------------------------------
;
; Now some clock supporting routines follow.
;
;-----------------------------------------------------------------------------

even

one_second_high		=	  91	;max count = 11931800, 2 seconds,
one_second_low		=	2124	; 327680 per tick, 5 per time
max_count_high		=	 182	; measuring unit of counter 0 of the
max_count_low		=	4248	; PIT, 59659 per 1/100 of a second

count_low		dw	0	;count is a double word
count_high		dw	0
timer_sync_switch	db	1	;is incremented during time setting
					; and decremented afterwards to avoid
					; unwanted interactions with automatic
					; synchronising
test_adjust_and_normalize_count proc near
	rcr	ah, 1			;test seconds, do we have the wrong
adjust_and_normalize_count proc near	; second?
	call	@@swap
	jnc	@@normalize
	add	ax, one_second_low	;here we adjust 1 second up
	adc	dx, one_second_high
@@normalize:				;adjust after updating of count
	cmp	dx, max_count_high	; after a roll-over max_count is
	jb	@@swap			; subtracted
	jg	@@adjust_down		; if count is too low max_count is
	ja	@@adjust_up		; added
	cmp	ax, max_count_low
	jb	@@swap
@@adjust_down:
	sub	ax, max_count_low
	sbb	dx, max_count_high
	jmp	@@normalize
@@adjust_up:
	add	ax, max_count_low
	adc	dx, max_count_high
	jmp	@@normalize
@@swap:	xchg	dx, [count_high]
	xchg	ax, [count_low]
	ret
adjust_and_normalize_count endp
test_adjust_and_normalize_count endp

;-----------------------------------------------------------------------------
;
; The timer interrupt (8) is intercepted. Every time it is called (18,2 times
; per second) we try to sync the clock.
;
;-----------------------------------------------------------------------------

timer	proc	far
	cli
	add	[count_high], 5		;this never sets the carry flag
	call	adjust_and_normalize_count
	cmp	[timer_sync_switch], 1
	jne	@@end_of_proc
	push	ax bx dx
	call	time_read_and_sync
	pop	dx bx ax
@@end_of_proc:

if	debug
	inc	[debug_timer_counter]
endif

	jmp_far	old_timer
timer	endp

;-----------------------------------------------------------------------------
;
; This routine reads the 1/100 of seconds from the Programmable Interrupt
; Timer, and synchronises when needed.
; If any routine sets the default rate of counter 0 of the PIT to something
; else than the default 65536, still calling the old Timer Interrupt 18.2
; times per second, then this all won't work properly (BASIC is known to do
; that in order to generate better sounds). It probably won't hang the
; computer, but the clock won't be accurate anymore to 1/100 of a second.
;
; The real syncing is done when the seconds change. The sync routine is called
; at every adjustment of the time, at every read of the time and 18.2 times
; per second. That should be enough to garantuee a precision that's more than
; 1/100 of a second.
;
;-----------------------------------------------------------------------------

time_read_and_sync proc near		;we'll read counter 0 of the PIT
	xor	bx, bx
	mov	dx, 5
	mov	al, 11100010b
	cli
	out	43h, al			;latch status of counter 0
	in	al, 40h			;read status
	xchg	bx, ax			;bl:= status, al:= 0
	out	70h, al			;now we read seconds from the RTC
	in	al, 71h
	call	bcd_bin_half
	xchg	bx, ax			;store seconds in bh, bl = unchanged
	mov	al, 00000110b
	out	43h, al			;latch count of counter 0
	in	al, 40h			;read least significant byte
	mov	ah, al
	in	al, 40h			;read most significant byte
	sti
	xchg	ah, al
	not	ax			;it was a count down, we go up
	cmp	bl, 80h			;no carry if OUT pin high
	rcr	ax, 1
	mul	dx
	add	ax, [count_low]
	adc	dx, [count_high]
	 push	ax
	mov	al, 11100010b
	cli
	out	43h, al			;reread status
	in	al, 40h			;if not the same, a change of status
	sti				; might have happened between reading
	cmp	al, bl			; the status and the count
	 pop	ax
if debug
	pushf
	je	end_write_s
	push	ax bx
	mov	ah, 0Eh
	mov	al, 's'			;write a 'S' to the screen
	mov	bx, 7
	int	10h
	pop	bx ax
	mov	[debug_S], 1
end_write_s:
	popf
endif
	jne	time_read_and_sync

tfhere	= this word
	push 	ax bx cx
	mov	ax, 160Ah		;MS Windows 3.1
	int	2Fh			; identify windows version and type
	or	ax, ax
	pop	cx bx ax
	jnz	@@end_win_ch
	mov	ax, [count_low]
	mov	dx, [count_high]
@@end_win_ch:
tfthere = this word

	push	ax dx			;store calculated count
	push	bx
	mov	bx, 59659
	div	bx			;get hundreds
	pop	bx
	aam	100			;al:= hundreds, ah:= seconds
	xor	ah, bh			;test wether last bits
	rcr	ah, 1			; are equal
	pop	dx			;get calculated count back
	jc	@@adjust
	pop	dx
	mov	ah, bh
if debug
	mov	[debug_S], 0
endif
	ret

@@adjust:

; By subtracting our calculated count from the master count, we set the actual
; count to 0.00.
; Under .50 we subtract one second and one unit more to get to 0.99
; After that we set the even/uneven second right.

if debug
	mov	[debug_100s], al
	jmp	@@debug_go_on
	db	'Debug timer counter:'
debug_timer_counter dw	0		;space for 1 hour
	db	'Debug sync counter:'
debug_sync_counter dw	0
debug_100s	db	0
debug_S		db	0
@@debug_go_on:
endif

	cmp	al, 50			;above or under .50?
	pop	ax			;get calculated count low
	mov	bl, 0			;up:   .50 -> 0.00
jmp_byte equ	this byte
	jmp	@@end_dec_swap		;we always move up at install time
; will become:
; 	jae	@@end_dec_swap
	mov	bl, 99			;down: <.50 -> 0.99
	sub	ax, one_second_low - 5	;1 second - 1 unit less down
	sbb	dx, one_second_high

if debug
	push	ax bx			;down: <.50 -> 0.99
	mov	ax, 0E44h		;write a 'D' to the screen
	cmp	[debug_100s], 0
	jne	@@do_write_1		;write a 'd' if small adjustment
	mov	al, 'd'
@@do_write_1:
	mov	bx, 7
	int	10h
	cmp	[debug_s], 1
	jne	d_go_on_1
	mov	ah, 0Eh
	mov	al, 'S'
	mov	bx, 7
	int	10h
d_go_on_1:
	pop	bx ax
	inc	[debug_sync_counter]
	jmp	@@end_dec_swap_2
endif

@@end_dec_swap:

if debug
	push	ax bx			;up:   .50 -> 0.00
	mov	ax, 0E55h		;write a 'U' to the screen
	cmp	[debug_100s], 99
	jne	@@do_write_2
	mov	al, 'u'			;write a 'u' if small adjustment
@@do_write_2:
	mov	bx, 7
	int	10h
	cmp	[debug_s], 1
	jne	d_go_on_2
	mov	ah, 0Eh
	mov	al, 'S'
	mov	bx, 7
	int	10h
d_go_on_2:
	pop	bx ax
	inc	[debug_sync_counter]
@@end_dec_swap_2:
endif

	cli
	sub	[count_low], ax		;adjust count to give 0.00 or 0.99
	sbb	[count_high], dx
	mov	ah, bh			;and adjust some more, getting the
	call	test_adjust_and_normalize_count ; even/uneven second right
	xchg	ax, bx
	sti
if debug
	mov	[debug_S], 0
endif
	ret
time_read_and_sync endp

;-----------------------------------------------------------------------------
;
; Interrupt 1A is intercepted too to be able to sync the clock when a new time
; is set. The setting of the time is done with only a precision of 1 second.
; For practical purposes that should be enough. Setting the RTC clock with a
; better precision is probably possible, but doesn't seem neccessary to me.
; By the way, many RTC clocks are not so precise and can miss a second or more
; per month.
;
;-----------------------------------------------------------------------------

int_1A	proc	far
	cmp	ah, 3
	je	@@function_3		;time set?
	cmp	ah, 4
	je	@@function_4		;date read?
to_old_int_1A:
	jmp_far	old_int_1A		;if not: go to old interrupt

@@function_3:
	dec	[timer_sync_switch]	;disable automatic syncing by int 8
	push	ax bx dx
@@again:call	time_read_and_sync
	cmp	al, 99			;wait if a change of second is uphand
	je	@@again
	pop	dx bx
	xor	ah, dh			;are the seconds changed from even to
	cli				; uneven or from uneven to even?
	call	test_adjust_and_normalize_count ; adjust count
	pop	ax
	pushf
	push	cs
	call	to_old_int_1A		;call old interrupt
	inc	[timer_sync_switch]	;enable automatic syncing
	iret

; Now follows the simple patch for a BIOS with a millennium bug.
; It makes sure that the century byte read is correct.

@@function_4:
	pushf
	push	cs
	call	to_old_int_1A		;call old interrupt
	cmp	cx, 1980h
	jae	@@ready
	mov	ch, 20h			;works OK at least until 2080
@@ready:iret				; if the century byte can be set
int_1A	endp				; it works until 2099

;-----------------------------------------------------------------------------
;
; Install request
;
; An amelioration might be that AT_Clock would adjust the maximum count at
; install time. The divide factor 59659 would have to be adjusted then and the
; maximum count would have to be adjusted with the same factor times 200.
;
; Checking this for my computer shows that there's not much to gain this way.
; Measurements give an average (out of 5 measurements) of 59658.52 as divide
; factor.
;
;-----------------------------------------------------------------------------

.8086	;the device won't work on a XT, but trying to install it should,
	; although it's useless, not hang a XT

	assume	cs:@code, ds:@code, es:nothing, ss:@code

init	proc	far
	push	ax bx cx dx si di es ds bp
	push	cs
	pop	ds			;point to our data segment
	pushf
	mov	[old_ss], ss
	mov	[old_sp], sp		;store stack
	mov	ax, cs
	cli				;set up new stack
	mov	ss, ax
	mov	sp, offset end_of_program
	mov	[pusha_instr], 60h	;overwrite jump to this init procedure
	mov	[push_ds_pushf], 9C1Eh
	sti

	mov	es, [db_ptr_s]		;point to device request header
	mov	si, [db_ptr_o]
	cld				;work upwards
	les	di, es:[si+cmdline]	;point to commandline in config.sys
	mov	al, '/'
	mov	cx, 128			;max length
@@cmdline_loop:
	cmp	byte ptr es:[di], CR	;ends with CR or LF
	je	@@end_of_cmdline_loop
	cmp	byte ptr es:[di], LF
	je	@@end_of_cmdline_loop
	scasb				;test for slash, preceeding any option
	jne	@@loop_again
	mov	bl, es:[di]
	or	bl, 20h
	cmp	bl, 'q'			;'/q' ?
	jne	@@check_u
	mov	[quiet], 1
	jmp	@@loop_again
@@check_u:
	cmp	bl, 'u'			;'/u' ?
	jne	@@set_help
	mov	[uninstall], 1
	jmp	@@loop_again
@@set_help:
	mov	[help], 1		;other options give help information
@@loop_again:
	loop	@@cmdline_loop
@@end_of_cmdline_loop:

	cmp	[help], 1
	jne	@@no_help
	mov	dx, offset help_txt	;help, don't install
	call	show_message
	jmp	@@init_ready
@@no_help:
	cmp	[uninstall], 1
	je	@@do_uninstall
	call	check_for_286
	cmp	ax, 1
	je	@@ok_286
	mov	dx, offset no_286m	;no 286+, don't install
@@show_message_and_abort:
	call	show_message
	mov	dx, offset not_inst_m
	call	show_message
	mov	dx, offset pause_m
	call	show_message
	mov	ah, 0
	int	16h
	mov	dx, offset new_lines
	call	show_message
	jmp	@@init_ready
@@ok_286:

	call	check_for_extended_aam_and_aad
	cmp	ax, 1
	je	@@ok_aam_aad
	mov	dx, offset no_xaamaad	;no extended AAM and AAD instructions
	jmp	@@show_message_and_abort; supported, don't install
@@ok_aam_aad:

	call	check_for_386
	cmp	ax, 1
	je	@@end_386		;put a jmp short instruction in to
	mov	tfhere, (((tfthere-tfhere)-2) shl 8) + 0EBh ; skip some code
@@end_386:

	clc
	mov	ah, 2
	int	1Ah
	jc	@@clock_not_operating
	mov	ch, 0
	mov	ah, 4
	int	1Ah
	jc	@@clock_not_operating
	cmp	ch, 19h
	je	@@ok_rtc
	cmp	ch, 20h
	je	@@ok_rtc
@@clock_not_operating:
	mov	dx, offset no_rtc	;BIOS says RTC not operating,
	jmp	@@show_message_and_abort; don't install
@@ok_RTC:

	mov	al, 0Dh
	out	70h, al
	in	al, 71h
	test	al, 80h			;CMOS RAM has lost power?
	jnz	@@ok_power_1
	mov	dx, offset no_power_1
	jmp	@@show_message_and_abort
@@ok_power_1:

	cmp	[quiet], 1		;show init message
	je	@@end_init_mess
	mov	dx, offset initm
	call	show_message
@@end_init_mess:

	mov	ax, 3508h
	int	21h
	mov	[old_timer.o], bx
	mov	[old_timer.s], es
	mov	dx, offset timer	;install timer
	mov	ax, 2508h
	int	21h
	mov	ax, 351Ah
	int	21h
	mov	[old_int_1A.o], bx
	mov	[old_int_1A.s], es
	mov	dx, offset int_1A	;install new get/set time interrupt
	mov	ax, 251Ah
	int	21h
					;and synchronise it
	call	time_read_and_sync
	mov	di, ax
@@sec_change_loop:			; making sure we have
	call	time_read_and_sync	; passed 2 second shifts, giving good
	mov	bx, di			; synchronisation
	cmp	bh, ah
	je	@@sec_change_loop
	mov	[jmp_byte], 73h		;introduce the possibility of
					; adjusting down
	mov	[res_size], offset init

; Some BIOS's can't make the century roll-over on their own, we check whether
; the BIOS missed one. If it did, we set the date again assuming that the
; BIOS does not have difficulty setting the century register of the RTC.

	mov	ah, 4			;read date
	int	1Ah
	cmp	ch, 19h			;it will help the 19-20 roll-over,
	ja	@@date_ok		; but not the 20-21 roll-over
	cmp	cl, 80h
	jae	@@date_ok
@@day_shift_loop_1:
	mov	ah, 2			;check wether a day shift is uphand
	int	1Ah			; and wait if it is
	cmp	cx, 2359h
	jne	@@end_of_day_shift_loop_1
	cmp	dh, 59h
	je	@@day_shift_loop_1
@@end_of_day_shift_loop_1:
	mov	ah, 4			;read date
	int	1Ah
	mov	ch, 20h
	mov	ah, 5			;set date
	int	1Ah			; thus setting the right century
@@date_ok:

@@init_ready:				;return
	mov	es, [db_ptr_s]
	mov	si, [db_ptr_o]

	mov	ax, [res_size]
	mov	es:[si+end_adr.o], ax	;end of resident code
	mov	es:[si+end_adr.s], cs
	mov	ax, [dev_status]
	mov	es:[si+status], ax	;signal ready

	cli
	mov	ss, [old_ss]
	mov	sp, [old_sp]		;restore stack
	sti
	popf
	pop	bp ds es di si dx cx bx ax
	ret

@@do_uninstall:
	mov	ax, 3508h
	int	21h
	cmp	bx, offset timer	;are the offsets OK
	jne	@@cant_uninstall
	mov	dx, es
	mov	ax, 351Ah
	int	21h
	mov	cx, es
	cmp	dx, cx
	jne	@@cant_uninstall	;"
	cmp	bx, offset int_1A
	jne	@@cant_uninstall
@@can_uninstall:
	push	es
	dec	dx
	mov	es, dx
	cmp	byte ptr [es:0], 'M'	;do we have a normal memory block
	je	@@ui_mb_ok		; so we are installed as 'TSR'
	cmp	byte ptr [es:0], 'Z'
	jne	@@cant_uninstall
@@ui_mb_ok:
	pop	es
	push	ds
	lds	dx, es:[old_timer]	;reset vectors
	mov	ax, 2508h
	int	21h
	lds	dx, es:[old_int_1A]	;"
	mov	ax, 251Ah
	int	21h
	pop	ds
	cmp	[quiet], 1
	je	@@init_ready
	mov	dx, offset uninstm	;say ready
	call	show_message		;rest is done by exe-part
	jmp	@@init_ready		; (unlinking it from the device driver
					;  chain, letting DOS point to another
					;  clock driver, releasing the memory
					;  block)
@@cant_uninstall:
	mov	dx, offset nouninstm
	call	show_message
	mov	[dev_status], 810Ch	;say: general failure
	jmp	@@init_ready
init	endp

show_message	proc			;shows message
	mov	ah, 9
	int	21h
	ret
show_message	endp

check_for_286	proc
; Checks for a 286 or better, ax holds result
	 pushf
	 pop	ax
	and	ax, 0FFFh		;if we can get bits 12-15
	 push	ax			; of the flags register
	 popf				; cleared, it's a 286+
	 pushf
	 pop	ax			;if not Fxxx, it's a 286+
	and	ax, 0F000h
	cmp	ax, 0F000h
	mov	ax, 0
	jz	@@end_of_proc
	inc	ax
@@end_of_proc:
	ret
check_for_286	endp

check_for_386	proc
; Checks for a 386 or better, ax holds result
	 pushf
	 pop	ax
	or	ax,3000h
	 push	ax
	 popf
	 pushf
	 pop	ax
	test	ax,3000h
	mov	ax, 0
	jz	@@end_of_proc
	inc	ax
@@end_of_proc:
	ret
check_for_386	endp

check_for_extended_AAM_and_AAD proc
; Are extended versions of AAM and AAD supported?
;
; See Hamarsoft's 86BUGS list, (C) 1993/94 By Hamarsoft (R)
;
; Testing the information in that list on a Nec V20, I found out that it's right:
; the extended version of AAM is supported by the NEC V20,
; but not the extended version of AAD: the NEC V20 always performs AAD 10d.
	mov	ax, 1234h
	aad	4
	cmp	ax, 7Ch
	jne	@@not_good
	mov	ax, 1234h
	aad	1
	cmp	ax, 46h
	jne	@@not_good
	mov	ax, 1234h
	aad	100d
	cmp	ax, 3Ch
	jne	@@not_good
	mov	ax, 1234h
	aam	4
	cmp	ax, 0D00h
	jne	@@not_good
	mov	ax, 1234h
	aam	1
	cmp	ax, 3400h
	jne	@@not_good
	mov	ax, 12F3h
	aam	100d
	cmp	ax, 22Bh
	jne	@@not_good
	mov	ax, 1
	ret
@@not_good:
	xor	ax, ax
	ret
check_for_extended_AAM_and_AAD endp

dataseg
help	db	0			;boolean
quiet	db	0			;boolean
uninstall db	0			;boolean
dev_status dw	100h			;presumed drive status
res_size dw	0h			;presume no residency

const
signature db 'by J. Martin Roelofs, ',??time,' ',??date
initm	db CR,LF,"J. Martin's AT clock driver ",version_nr,' installed.',CR,LF,'$'
uninstm	db CR,LF,"J. Martin's AT clock driver ",version_nr,' uninstalled.',CR,LF,'$'
nouninstm db CR,LF,"Cannot uninstall J. Martin's AT clock driver ",version_nr,'.',CR,LF,'$'
not_inst_m db ",",CR,LF,"J. Martin's AT clock driver ",version_nr,' not installed.',CR,LF,'$'
no_286m	db CR,LF,'This is no AT+','$'
no_xaamaad db CR,LF,'Extended AAM and AAD instructions not supported','$'
no_rtc	db CR,LF,'No operating Real Time Clock found','$'
no_power_1 db CR,LF,'CMOS RAM has lost power','$'
pause_m db 'Strike any key to continue...','$'
new_lines db CR,LF,CR,LF,'$'

help_txt db CR,LF
	db "J. Martin's AT clock driver ",version_nr,' information.',CR,LF
	db CR,LF
	db "This is a device driver that make's DOS always read the AT Real Time Clock",CR,LF
	db 'for date and time functions. It also uses the Programmable Interrupt Timer.',CR,LF
	db "It functions with a precision of 1/100 of a second, is millennium proof,",CR,LF
	db "patches the BIOS century roll-over bug, and takes care of the date roll-over",CR,LF
	db 'problem. It can also be installed as TSR.',CR,LF
	db CR,LF
	db '    device=[drive:][path]at_clock.exe [/q]',CR,LF
	db CR,LF
	db '    at_clock [/q] [/u]',CR,LF
	db CR,LF
	db '    /q  quiet, no messages',CR,LF
	db '    /u  uninstall (only if installed as TSR)',CR,LF
	db CR,LF,'$'

udataseg
old_sp	dw	?			;storage place for old stack
old_ss	dw	?
even
our_stack db	280h dup (?)

end_of_program:

end